@@ -10,30 +10,29 @@
 #![cfg_attr(docsrs, doc(cfg(unstable)))]
 #![doc(html_logo_url = "https://knurling.ferrous-systems.com/knurling_logo_light_text.svg")]
 
-// load DEFMT_VERSION
-include!(concat!(env!("OUT_DIR"), "/version.rs"));
+pub const DEFMT_VERSION: &str = "3";
 
 mod decoder;
 mod elf2table;
 mod frame;
 pub mod log;
+mod stream;
 
 use std::{
-    collections::BTreeMap,
+    collections::{BTreeMap, HashMap},
     error::Error,
     fmt, io,
-    sync::{
-        atomic::{self, AtomicBool},
-        Arc,
-    },
+    str::FromStr,
 };
 
-use decoder::{read_leb128, Decoder};
+use byteorder::{ReadBytesExt, LE};
+use decoder::Decoder;
 use defmt_parser::Level;
 use elf2table::parse_impl;
 
 pub use elf2table::{Location, Locations};
 pub use frame::Frame;
+pub use stream::StreamDecoder;
 
 /// Specifies the origin of a format string
 #[derive(PartialEq, Eq, Debug)]
@@ -42,6 +41,8 @@ pub enum Tag {
     Prim,
     /// Format string created by `#[derive(Format)]`.
     Derived,
+    /// Format string created by `defmt::bitflags!`.
+    Bitflags,
     /// A user-defined format string from a `write!` invocation.
     Write,
     /// An interned string, for use with `{=istr}`.
@@ -49,6 +50,11 @@ pub enum Tag {
     /// Defines the global timestamp format.
     Timestamp,
 
+    /// `static` containing a possible value of a bitflags type.
+    BitflagsValue,
+    /// Format string created by `defmt::println!`.
+    Println,
+
     Trace,
     Debug,
     Info,
@@ -69,8 +75,8 @@ impl Tag {
     }
 }
 
-/// Entry in [`Table`] combining a format string with it's raw symbol
-#[derive(Debug)]
+/// Entry in [`Table`] combining a format string with its raw symbol
+#[derive(Debug, PartialEq)]
 pub struct TableEntry {
     string: StringEntry,
     raw_symbol: String,
@@ -91,7 +97,7 @@ impl TableEntry {
 }
 
 /// A format string and it's [`Tag`]
-#[derive(Debug)]
+#[derive(Debug, PartialEq)]
 pub struct StringEntry {
     tag: Tag,
     string: String,
@@ -103,22 +109,53 @@ impl StringEntry {
     }
 }
 
+/// Data that uniquely identifies a `defmt::bitflags!` invocation.
+#[derive(Debug, PartialEq, Eq, Hash)]
+struct BitflagsKey {
+    /// Name of the bitflags struct (this is really redundant with `disambig`).
+    ident: String,
+    package: String,
+    disambig: String,
+}
+
+#[derive(Copy, Clone, Debug, PartialEq)]
+#[non_exhaustive]
+pub enum Encoding {
+    Raw,
+    Rzcobs,
+}
+
+impl FromStr for Encoding {
+    type Err = anyhow::Error;
+
+    fn from_str(s: &str) -> Result<Self, Self::Err> {
+        match s {
+            "raw" => Ok(Encoding::Raw),
+            "rzcobs" => Ok(Encoding::Rzcobs),
+            _ => anyhow::bail!("Unknown defmt encoding '{}' specified. This is a bug.", s),
+        }
+    }
+}
+
+impl Encoding {
+    pub const fn can_recover(&self) -> bool {
+        match self {
+            Encoding::Raw => false,
+            Encoding::Rzcobs => true,
+        }
+    }
+}
+
 /// Internal table that holds log levels and maps format strings to indices
-#[derive(Debug)]
+#[derive(Debug, PartialEq)]
 pub struct Table {
     timestamp: Option<TableEntry>,
     entries: BTreeMap<usize, TableEntry>,
+    bitflags: HashMap<BitflagsKey, Vec<(String, u128)>>,
+    encoding: Encoding,
 }
 
 impl Table {
-    /// NOTE caller must verify that defmt symbols are compatible with this version of the `decoder` crate using the `check_version` function
-    pub fn new(entries: BTreeMap<usize, TableEntry>) -> Self {
-        Self {
-            entries,
-            timestamp: None,
-        }
-    }
-
     /// Parses an ELF file and returns the decoded `defmt` table.
     ///
     /// This function returns `None` if the ELF file contains no `.defmt` section.
@@ -142,9 +179,8 @@ impl Table {
         Ok((entry.string.tag.to_level(), &entry.string.string))
     }
 
-    fn get_with_level(&self, index: usize) -> Result<(Level, &str), ()> {
-        let (lvl, format) = self._get(index)?;
-        Ok((lvl.ok_or(())?, format))
+    fn get_with_level(&self, index: usize) -> Result<(Option<Level>, &str), ()> {
+        self._get(index)
     }
 
     fn get_without_level(&self, index: usize) -> Result<&str, ()> {
@@ -158,7 +194,7 @@ impl Table {
 
     pub fn indices(&self) -> impl Iterator<Item = usize> + '_ {
         self.entries.iter().filter_map(move |(idx, entry)| {
-            if entry.string.tag.to_level().is_some() {
+            if entry.string.tag.to_level().is_some() || entry.string.tag == Tag::Println {
                 Some(*idx)
             } else {
                 None
@@ -179,16 +215,17 @@ impl Table {
         elf2table::get_locations(elf, self)
     }
 
-    /// decode the data sent by the device using the previosuly stored metadata
+    /// Decode the data sent by the device using the previously stored metadata.
     ///
-    /// * bytes: contains the data sent by the device that logs.
-    ///          contains the [log string index, timestamp, optional fmt string args]
+    /// * `bytes`
+    ///   * contains the data sent by the device that logs.
+    ///   * contains the [log string index, timestamp, optional fmt string args]
     pub fn decode<'t>(
         &'t self,
         mut bytes: &[u8],
-    ) -> Result<(Frame<'t>, /*consumed: */ usize), DecodeError> {
+    ) -> Result<(Frame<'t>, /* consumed: */ usize), DecodeError> {
         let len = bytes.len();
-        let index = read_leb128(&mut bytes)?;
+        let index = bytes.read_u16::<LE>()? as u64;
 
         let mut decoder = Decoder::new(self, bytes);
 
@@ -205,54 +242,43 @@ impl Table {
             .map_err(|_| DecodeError::Malformed)?;
 
         let args = decoder.decode_format(format)?;
-        if !decoder.bools_tbd.is_empty() {
-            // Flush end of compression block.
-            decoder.read_and_unpack_bools()?;
-        }
 
-        let frame = Frame::new(level, index, timestamp_format, timestamp_args, format, args);
+        let frame = Frame::new(
+            self,
+            level,
+            index,
+            timestamp_format,
+            timestamp_args,
+            format,
+            args,
+        );
 
         let consumed = len - decoder.bytes.len();
         Ok((frame, consumed))
     }
-}
 
-#[derive(Debug)]
-struct Bool(AtomicBool);
-
-impl Bool {
-    #[allow(clippy::declare_interior_mutable_const)]
-    const FALSE: Self = Self(AtomicBool::new(false));
-
-    fn set(&self, value: bool) {
-        self.0.store(value, atomic::Ordering::Relaxed);
+    pub fn new_stream_decoder(&self) -> Box<dyn StreamDecoder + '_> {
+        match self.encoding {
+            Encoding::Raw => Box::new(stream::Raw::new(self)),
+            Encoding::Rzcobs => Box::new(stream::Rzcobs::new(self)),
+        }
     }
-}
 
-impl fmt::Display for Bool {
-    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        write!(f, "{:?}", self.0.load(atomic::Ordering::Relaxed))
-    }
-}
-
-impl PartialEq for Bool {
-    fn eq(&self, other: &Self) -> bool {
-        self.0
-            .load(atomic::Ordering::Relaxed)
-            .eq(&other.0.load(atomic::Ordering::Relaxed))
+    pub fn encoding(&self) -> Encoding {
+        self.encoding
     }
 }
 
 // NOTE follows `parser::Type`
-#[derive(Debug, PartialEq)]
+#[derive(Debug, Clone, PartialEq)]
 enum Arg<'t> {
     /// Bool
-    Bool(Arc<Bool>),
+    Bool(bool),
     F32(f32),
     F64(f64),
-    /// U8, U16, U24 and U32
+    /// U8, U16, U32, U64, U128
     Uxx(u128),
-    /// I8, I16, I24 and I32
+    /// I8, I16, I32, I64, I128
     Ixx(i128),
     /// Str
     Str(String),
@@ -266,6 +292,9 @@ enum Arg<'t> {
     FormatSlice {
         elements: Vec<FormatSliceElement<'t>>,
     },
+    FormatSequence {
+        args: Vec<Arg<'t>>,
+    },
     /// Slice or Array of bytes.
     Slice(Vec<u8>),
     /// Char
@@ -275,7 +304,7 @@ enum Arg<'t> {
     Preformatted(String),
 }
 
-#[derive(Debug, PartialEq)]
+#[derive(Debug, Clone, PartialEq)]
 struct FormatSliceElement<'t> {
     // this will usually be the same format string for all elements; except when the format string
     // is an enum -- in that case `format` will be the variant
@@ -316,6 +345,30 @@ impl Error for DecodeError {}
 mod tests {
     use super::*;
 
+    fn test_table(entries: impl IntoIterator<Item = TableEntry>) -> Table {
+        Table {
+            timestamp: None,
+            entries: entries.into_iter().enumerate().collect(),
+            bitflags: Default::default(),
+            encoding: Encoding::Raw,
+        }
+    }
+
+    fn test_table_with_timestamp(
+        entries: impl IntoIterator<Item = TableEntry>,
+        timestamp: &str,
+    ) -> Table {
+        Table {
+            timestamp: Some(TableEntry::new_without_symbol(
+                Tag::Timestamp,
+                timestamp.into(),
+            )),
+            entries: entries.into_iter().enumerate().collect(),
+            bitflags: Default::default(),
+            encoding: Encoding::Raw,
+        }
+    }
+
     // helper function to initiate decoding and assert that the result is as expected.
     //
     // format:       format string to be expanded
@@ -332,8 +385,10 @@ mod tests {
             entries,
             timestamp: Some(TableEntry::new_without_symbol(
                 Tag::Timestamp,
-                "{=u8:µs}".to_owned(),
+                "{=u8:us}".to_owned(),
             )),
+            bitflags: Default::default(),
+            encoding: Encoding::Raw,
         };
 
         let frame = table.decode(&bytes).unwrap().0;
@@ -342,37 +397,34 @@ mod tests {
 
     #[test]
     fn decode() {
-        let mut entries = BTreeMap::new();
-        entries.insert(
-            0,
+        let entries = vec![
             TableEntry::new_without_symbol(Tag::Info, "Hello, world!".to_owned()),
-        );
-        entries.insert(
-            1,
             TableEntry::new_without_symbol(Tag::Debug, "The answer is {=u8}!".to_owned()),
-        );
-        // [IDX, TS, 42]
-        //           ^^
-        //entries.insert(2, "The answer is {0:u8} {1:u16}!".to_owned());
+        ];
 
-        let table = Table {
-            entries,
-            timestamp: None,
-        };
+        let table = test_table(entries);
 
-        let bytes = [0];
+        let bytes = [0, 0];
         //     index ^
 
         assert_eq!(
             table.decode(&bytes),
             Ok((
-                Frame::new(Level::Info, 0, None, vec![], "Hello, world!", vec![],),
+                Frame::new(
+                    &table,
+                    Some(Level::Info),
+                    0,
+                    None,
+                    vec![],
+                    "Hello, world!",
+                    vec![],
+                ),
                 bytes.len(),
             ))
         );
 
         let bytes = [
-            1,  // index
+            1, 0,  // index
             42, // argument
         ];
 
@@ -380,7 +432,8 @@ mod tests {
             table.decode(&bytes),
             Ok((
                 Frame::new(
-                    Level::Debug,
+                    &table,
+                    Some(Level::Debug),
                     1,
                     None,
                     vec![],
@@ -397,20 +450,16 @@ mod tests {
     #[test]
     fn all_integers() {
         const FMT: &str =
-            "Hello, {=u8} {=u16} {=u24} {=u32} {=u64} {=u128} {=i8} {=i16} {=i32} {=i64} {=i128}!";
-        let mut entries = BTreeMap::new();
-        entries.insert(0, TableEntry::new_without_symbol(Tag::Info, FMT.to_owned()));
+            "Hello, {=u8} {=u16} {=u32} {=u64} {=u128} {=i8} {=i16} {=i32} {=i64} {=i128}!";
 
-        let table = Table {
-            entries,
-            timestamp: None,
-        };
+        let entries = vec![TableEntry::new_without_symbol(Tag::Info, FMT.to_owned())];
+
+        let table = test_table(entries);
 
         let bytes = [
-            0,  // index
+            0, 0,  // index
             42, // u8
             0xff, 0xff, // u16
-            0, 0, 1, // u24
             0xff, 0xff, 0xff, 0xff, // u32
             0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, // u64
             0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
@@ -427,7 +476,8 @@ mod tests {
             table.decode(&bytes),
             Ok((
                 Frame::new(
-                    Level::Info,
+                    &table,
+                    Some(Level::Info),
                     0,
                     None,
                     vec![],
@@ -435,7 +485,6 @@ mod tests {
                     vec![
                         Arg::Uxx(42),                      // u8
                         Arg::Uxx(u16::max_value().into()), // u16
-                        Arg::Uxx(0x10000),                 // u24
                         Arg::Uxx(u32::max_value().into()), // u32
                         Arg::Uxx(u64::max_value().into()), // u64
                         Arg::Uxx(u128::max_value()),       // u128
@@ -453,25 +502,17 @@ mod tests {
 
     #[test]
     fn indices() {
-        let mut entries = BTreeMap::new();
-        entries.insert(
-            0,
+        let entries = vec![
             TableEntry::new_without_symbol(Tag::Info, "The answer is {0=u8} {0=u8}!".to_owned()),
-        );
-        entries.insert(
-            1,
             TableEntry::new_without_symbol(
                 Tag::Info,
                 "The answer is {1=u16} {0=u8} {1=u16}!".to_owned(),
             ),
-        );
+        ];
 
-        let table = Table {
-            entries,
-            timestamp: None,
-        };
+        let table = test_table(entries);
         let bytes = [
-            0,  // index
+            0, 0,  // index
             42, // argument
         ];
 
@@ -479,7 +520,8 @@ mod tests {
             table.decode(&bytes),
             Ok((
                 Frame::new(
-                    Level::Info,
+                    &table,
+                    Some(Level::Info),
                     0,
                     None,
                     vec![],
@@ -491,7 +533,7 @@ mod tests {
         );
 
         let bytes = [
-            1,  // index
+            1, 0,  // index
             42, // u8
             0xff, 0xff, // u16
         ];
@@ -500,7 +542,8 @@ mod tests {
             table.decode(&bytes),
             Ok((
                 Frame::new(
-                    Level::Info,
+                    &table,
+                    Some(Level::Info),
                     1,
                     None,
                     vec![],
@@ -514,24 +557,16 @@ mod tests {
 
     #[test]
     fn format() {
-        let mut entries = BTreeMap::new();
-        entries.insert(
-            0,
+        let entries = vec![
             TableEntry::new_without_symbol(Tag::Info, "x={=?}".to_owned()),
-        );
-        entries.insert(
-            1,
             TableEntry::new_without_symbol(Tag::Derived, "Foo {{ x: {=u8} }}".to_owned()),
-        );
+        ];
 
-        let table = Table {
-            entries,
-            timestamp: None,
-        };
+        let table = test_table(entries);
 
         let bytes = [
-            0,  // index
-            1,  // index of the struct
+            0, 0, // index
+            1, 0,  // index of the struct
             42, // Foo.x
         ];
 
@@ -539,7 +574,8 @@ mod tests {
             table.decode(&bytes),
             Ok((
                 Frame::new(
-                    Level::Info,
+                    &table,
+                    Some(Level::Info),
                     0,
                     None,
                     vec![],
@@ -555,29 +591,71 @@ mod tests {
     }
 
     #[test]
-    fn display() {
-        let mut entries = BTreeMap::new();
-        entries.insert(
-            0,
-            TableEntry::new_without_symbol(Tag::Info, "x={=?}".to_owned()),
-        );
-        entries.insert(
-            1,
-            TableEntry::new_without_symbol(Tag::Derived, "Foo {{ x: {=u8} }}".to_owned()),
-        );
+    fn format_sequence() {
+        let entries = vec![
+            TableEntry::new_without_symbol(Tag::Info, "{=__internal_FormatSequence}".to_owned()),
+            TableEntry::new_without_symbol(Tag::Derived, "Foo".to_owned()),
+            TableEntry::new_without_symbol(Tag::Derived, "Bar({=u8})".to_owned()),
+            TableEntry::new_without_symbol(Tag::Derived, "State {=u8}|".to_owned()),
+        ];
 
-        let table = Table {
-            entries,
-            timestamp: Some(TableEntry::new_without_symbol(
-                Tag::Timestamp,
-                "{=u8:µs}".to_owned(),
-            )),
-        };
+        let table = test_table(entries);
 
         let bytes = [
-            0,  // index
-            2,  // timestamp
-            1,  // index of the struct
+            0, 0, // index
+            1, 0, // index of Foo
+            2, 0,  // index of Bar
+            42, // bar.x
+            3, 0,  // index of State
+            23, // State variable
+            0, 0, // terminator
+        ];
+
+        assert_eq!(
+            table.decode(&bytes),
+            Ok((
+                Frame::new(
+                    &table,
+                    Some(Level::Info),
+                    0,
+                    None,
+                    vec![],
+                    "{=__internal_FormatSequence}",
+                    vec![Arg::FormatSequence {
+                        args: vec![
+                            Arg::Format {
+                                format: "Foo",
+                                args: vec![]
+                            },
+                            Arg::Format {
+                                format: "Bar({=u8})",
+                                args: vec![Arg::Uxx(42)]
+                            },
+                            Arg::Format {
+                                format: "State {=u8}|",
+                                args: vec![Arg::Uxx(23)]
+                            }
+                        ]
+                    }],
+                ),
+                bytes.len(),
+            ))
+        );
+    }
+
+    #[test]
+    fn display() {
+        let entries = vec![
+            TableEntry::new_without_symbol(Tag::Info, "x={=?}".to_owned()),
+            TableEntry::new_without_symbol(Tag::Derived, "Foo {{ x: {=u8} }}".to_owned()),
+        ];
+
+        let table = test_table_with_timestamp(entries, "{=u8:us}");
+
+        let bytes = [
+            0, 0, // index
+            2, // timestamp
+            1, 0,  // index of the struct
             42, // Foo.x
         ];
 
@@ -588,10 +666,123 @@ mod tests {
         );
     }
 
+    #[test]
+    fn display_use_inner_type_hint() {
+        let entries = vec![
+            TableEntry::new_without_symbol(Tag::Info, "x={:b}".to_owned()),
+            TableEntry::new_without_symbol(Tag::Derived, "S {{ x: {=u8:x} }}".to_owned()),
+        ];
+
+        let table = test_table_with_timestamp(entries, "{=u8:us}");
+
+        let bytes = [
+            0, 0, // index
+            2, // timestamp
+            1, 0,  // index of the struct
+            42, // value
+        ];
+
+        let frame = table.decode(&bytes).unwrap().0;
+        assert_eq!(
+            frame.display(false).to_string(),
+            "0.000002 INFO x=S { x: 2a }",
+        );
+    }
+
+    #[test]
+    fn display_use_outer_type_hint() {
+        let entries = vec![
+            TableEntry::new_without_symbol(Tag::Info, "x={:b}".to_owned()),
+            TableEntry::new_without_symbol(Tag::Derived, "S {{ x: {=u8:?} }}".to_owned()),
+        ];
+
+        let table = test_table_with_timestamp(entries, "{=u8:us}");
+
+        let bytes = [
+            0, 0, // index
+            2, // timestamp
+            1, 0,  // index of the struct
+            42, // value
+        ];
+
+        let frame = table.decode(&bytes).unwrap().0;
+        assert_eq!(
+            frame.display(false).to_string(),
+            "0.000002 INFO x=S { x: 101010 }",
+        );
+    }
+
+    #[test]
+    fn display_inner_str_in_struct() {
+        let entries = vec![
+            TableEntry::new_without_symbol(Tag::Info, "{}".to_owned()),
+            TableEntry::new_without_symbol(Tag::Derived, "S {{ x: {=str:?} }}".to_owned()),
+        ];
+
+        let table = test_table_with_timestamp(entries, "{=u8:us}");
+
+        let bytes = [
+            0, 0, // index
+            2, // timestamp
+            1, 0, // index into the struct
+            5, 0, 0, 0, // length of the string
+            b'H', b'e', b'l', b'l', b'o', // string "Hello"
+        ];
+        let frame = table.decode(&bytes).unwrap().0;
+        assert_eq!(
+            frame.display(false).to_string(),
+            "0.000002 INFO S { x: \"Hello\" }",
+        );
+    }
+
+    #[test]
+    fn display_u8_vec() {
+        let entries = vec![
+            TableEntry::new_without_symbol(Tag::Prim, "{=u8}".to_owned()),
+            TableEntry::new_without_symbol(Tag::Prim, "{=[?]}".to_owned()),
+            TableEntry::new_without_symbol(Tag::Derived, "Data {{ name: {=?:?} }}".to_owned()),
+            TableEntry::new_without_symbol(Tag::Info, "{=[?]:a}".to_owned()),
+        ];
+
+        let table = test_table_with_timestamp(entries, "{=u8:us}");
+
+        let bytes = [
+            3, 0, // frame index
+            2, // timestamp value of type `u8`
+            1, 0, 0, 0, // number of elements in `FormatSlice`
+            2, 0, // index to `Data` struct
+            1, 0, // Format index to table entry: `{=[?]}`
+            2, 0, 0, 0, // inner FormatSlice, number of elements in `name` field
+            0, 0,   // Format index to table entry: `{=u8}`
+            72,  // "H"
+            105, // "i"
+        ];
+        let frame = table.decode(&bytes).unwrap().0;
+        assert_eq!(
+            frame.display(false).to_string(),
+            "0.000002 INFO [Data { name: b\"Hi\" }]",
+        );
+    }
+
+    #[test]
+    fn display_iso8601_timestamp() {
+        let bytes = [
+            0, 0, // index
+            2, // timestamp
+            36, 188, 151, 238, 120, 1, 0, 0, // unix timestamp in bytes: 1618910624804
+        ];
+
+        decode_and_expect(
+            "{=u64:iso8601ms}",
+            &bytes,
+            "0.000002 INFO 2021-04-20T09:23:44.804Z",
+        );
+    }
+
     #[test]
     fn bools_simple() {
         let bytes = [
-            0,          // index
+            0, 0,          // index
             2,          // timestamp
             true as u8, // the logged bool value
         ];
@@ -599,153 +790,10 @@ mod tests {
         decode_and_expect("my bool={=bool}", &bytes, "0.000002 INFO my bool=true");
     }
 
-    #[test]
-    fn bools_max_capacity() {
-        let bytes = [
-            0,           // index
-            2,           // timestamp
-            0b0110_0001, // the first 8 logged bool values
-        ];
-
-        decode_and_expect(
-            "bool capacity {=bool} {=bool} {=bool} {=bool} {=bool} {=bool} {=bool} {=bool}",
-            &bytes,
-            "0.000002 INFO bool capacity false true true false false false false true",
-        );
-    }
-
-    #[test]
-    fn bools_more_than_fit_in_one_byte() {
-        let bytes = [
-            0,           // index
-            2,           // timestamp
-            0b0110_0001, // the first 8 logged bool values
-            0b1,         // the final logged bool value
-        ];
-
-        decode_and_expect(
-            "bool overflow {=bool} {=bool} {=bool} {=bool} {=bool} {=bool} {=bool} {=bool} {=bool}",
-            &bytes,
-            "0.000002 INFO bool overflow false true true false false false false true true",
-        );
-
-        // Ensure that bools are compressed into the first byte even when there's non-bool values
-        // between them.
-        let bytes = [
-            0,           // index
-            2,           // timestamp
-            0xff,        // the logged u8
-            0b0110_0001, // the first 8 logged bool values
-            0b1,         // the final logged bool value
-        ];
-
-        decode_and_expect(
-            "bool overflow {=bool} {=u8} {=bool} {=bool} {=bool} {=bool} {=bool} {=bool} {=bool} {=bool}",
-            &bytes,
-            "0.000002 INFO bool overflow false 255 true true false false false false true true",
-        );
-
-        // Ensure that bools are compressed into the first byte even when there's a non-bool value
-        // right between between the two compression blocks.
-        let bytes = [
-            0,           // index
-            2,           // timestamp
-            0b0110_0001, // the first 8 logged bool values
-            0xff,        // the logged u8
-            0b1,         // the final logged bool value
-        ];
-
-        decode_and_expect(
-            "bool overflow {=bool} {=bool} {=bool} {=bool} {=bool} {=bool} {=bool} {=bool} {=u8} {=bool}",
-            &bytes,
-            "0.000002 INFO bool overflow false true true false false false false true 255 true",
-        );
-    }
-
-    #[test]
-    fn bools_mixed() {
-        let bytes = [
-            0,       // index
-            2,       // timestamp
-            9 as u8, // a uint in between
-            0b101,   // 3 packed bools
-        ];
-
-        decode_and_expect(
-            "hidden bools {=bool} {=u8} {=bool} {=bool}",
-            &bytes,
-            "0.000002 INFO hidden bools true 9 false true",
-        );
-    }
-
-    #[test]
-    fn bools_mixed_no_trailing_bool() {
-        let bytes = [
-            0,   // index
-            2,   // timestamp
-            9,   // a u8 in between
-            0b0, // 3 packed bools
-        ];
-
-        decode_and_expect(
-            "no trailing bools {=bool} {=u8}",
-            &bytes,
-            "0.000002 INFO no trailing bools false 9",
-        );
-    }
-
-    #[test]
-    fn bools_bool_struct() {
-        /*
-        emulate
-        #[derive(Format)]
-        struct Flags {
-            a: bool,
-            b: bool,
-            c: bool,
-        }
-
-        defmt::info!("{:bool} {:?}", true, Flags {a: true, b: false, c: true });
-        */
-
-        let mut entries = BTreeMap::new();
-        entries.insert(
-            0,
-            TableEntry::new_without_symbol(Tag::Info, "{=bool} {=?}".to_owned()),
-        );
-        entries.insert(
-            1,
-            TableEntry::new_without_symbol(
-                Tag::Derived,
-                "Flags {{ a: {=bool}, b: {=bool}, c: {=bool} }}".to_owned(),
-            ),
-        );
-
-        let table = Table {
-            entries,
-            timestamp: Some(TableEntry::new_without_symbol(
-                Tag::Timestamp,
-                "{=u8:µs}".to_owned(),
-            )),
-        };
-
-        let bytes = [
-            0,      // index
-            2,      // timestamp
-            1,      // index of Flags { a: {:bool}, b: {:bool}, c: {:bool} }
-            0b1101, // 4 packed bools
-        ];
-
-        let frame = table.decode(&bytes).unwrap().0;
-        assert_eq!(
-            frame.display(false).to_string(),
-            "0.000002 INFO true Flags { a: true, b: false, c: true }"
-        );
-    }
-
     #[test]
     fn bitfields() {
         let bytes = [
+            0,
             0,           // index
             2,           // timestamp
             0b1110_0101, // u8
@@ -760,6 +808,7 @@ mod tests {
     #[test]
     fn bitfields_reverse_order() {
         let bytes = [
+            0,
             0,           // index
             2,           // timestamp
             0b1101_0010, // u8
@@ -774,6 +823,7 @@ mod tests {
     #[test]
     fn bitfields_different_indices() {
         let bytes = [
+            0,
             0,           // index
             2,           // timestamp
             0b1111_0000, // u8
@@ -789,6 +839,7 @@ mod tests {
     #[test]
     fn bitfields_u16() {
         let bytes = [
+            0,
             0, // index
             2, // timestamp
             0b1111_0000,
@@ -800,6 +851,7 @@ mod tests {
     #[test]
     fn bitfields_mixed_types() {
         let bytes = [
+            0,
             0, // index
             2, // timestamp
             0b1111_0000,
@@ -816,6 +868,7 @@ mod tests {
     #[test]
     fn bitfields_mixed() {
         let bytes = [
+            0,
             0, // index
             2, // timestamp
             0b1111_0000,
@@ -833,6 +886,7 @@ mod tests {
     #[test]
     fn bitfields_across_boundaries() {
         let bytes = [
+            0,
             0, // index
             2, // timestamp
             0b1101_0010,
@@ -848,6 +902,7 @@ mod tests {
     #[test]
     fn bitfields_across_boundaries_diff_indices() {
         let bytes = [
+            0,
             0, // index
             2, // timestamp
             0b1101_0010,
@@ -864,6 +919,7 @@ mod tests {
     #[test]
     fn bitfields_truncated_front() {
         let bytes = [
+            0,
             0,           // index
             2,           // timestamp
             0b0110_0011, // truncated(!) u16
@@ -878,6 +934,7 @@ mod tests {
     #[test]
     fn bitfields_non_truncated_u32() {
         let bytes = [
+            0,
             0,           // index
             2,           // timestamp
             0b0110_0011, // -
@@ -895,6 +952,7 @@ mod tests {
     #[test]
     fn bitfields_u128() {
         let bytes = [
+            0,
             0,           // index
             2,           // timestamp
             0b1110_0101, // 120..127
@@ -920,9 +978,9 @@ mod tests {
     #[test]
     fn slice() {
         let bytes = [
-            0, // index
+            0, 0, // index
             2, // timestamp
-            2, // length of the slice
+            2, 0, 0, 0, // length of the slice
             23, 42, // slice content
         ];
         decode_and_expect("x={=[u8]}", &bytes, "0.000002 INFO x=[23, 42]");
@@ -931,9 +989,9 @@ mod tests {
     #[test]
     fn slice_with_trailing_args() {
         let bytes = [
-            0, // index
+            0, 0, // index
             2, // timestamp
-            2, // length of the slice
+            2, 0, 0, 0, // length of the slice
             23, 42, // slice content
             1,  // trailing arg
         ];
@@ -948,9 +1006,9 @@ mod tests {
     #[test]
     fn string_hello_world() {
         let bytes = [
-            0, // index
+            0, 0, // index
             2, // timestamp
-            5, // length of the string
+            5, 0, 0, 0, // length of the string
             b'W', b'o', b'r', b'l', b'd',
         ];
 
@@ -960,9 +1018,9 @@ mod tests {
     #[test]
     fn string_with_trailing_data() {
         let bytes = [
-            0, // index
+            0, 0, // index
             2, // timestamp
-            5, // length of the string
+            5, 0, 0, 0, // length of the string
             b'W', b'o', b'r', b'l', b'd', 125, // trailing data
         ];
 
@@ -976,7 +1034,7 @@ mod tests {
     #[test]
     fn char_data() {
         let bytes = [
-            0, // index
+            0, 0, // index
             2, // timestamp
             0x61, 0x00, 0x00, 0x00, // char 'a'
             0x9C, 0xF4, 0x01, 0x00, // Purple heart emoji
@@ -1009,16 +1067,18 @@ mod tests {
             entries,
             timestamp: Some(TableEntry::new_without_symbol(
                 Tag::Timestamp,
-                "{=u8:µs}".to_owned(),
+                "{=u8:us}".to_owned(),
             )),
+            bitflags: Default::default(),
+            encoding: Encoding::Raw,
         };
 
         let bytes = [
-            4,  // string index (INFO)
-            0,  // timestamp
-            3,  // string index (enum)
-            1,  // Some discriminant
-            2,  // string index (u8)
+            4, 0, // string index (INFO)
+            0, // timestamp
+            3, 0, // string index (enum)
+            1, // Some discriminant
+            2, 0,  // string index (u8)
             42, // Some.0
         ];
 
@@ -1026,9 +1086,9 @@ mod tests {
         assert_eq!(frame.display(false).to_string(), "0.000000 INFO x=Some(42)");
 
         let bytes = [
-            4, // string index (INFO)
+            4, 0, // string index (INFO)
             1, // timestamp
-            3, // string index (enum)
+            3, 0, // string index (enum)
             0, // None discriminant
         ];
 
